﻿#include <QAudioDeviceInfo>
#include <QDebug>
#include "../common/sources/system/filecache.h"

#include "sound.h"
#include "player.h"
#include "interface.h"

#define TIMER_PERIOD    17
#define SAMPLE_RATE     6000
#define CHANNELS        2
#define SAMPLE_SIZE     16
#define SAMPLE_BYTES    (SAMPLE_SIZE / 8)
#define MIXER_COEF      4   // nombre de sons simultanés
#define WAV_HEADER      44  // taille de l'entête d'un fichier ".wav"

//------------------------------------------------------------
CSoundManager::CSoundManager()
{
    memset(mBuffer, 0, SND_BUFFER_SIZE);

    QAudioFormat    format;
    format.setSampleRate(SAMPLE_RATE);
    format.setChannelCount(CHANNELS);
    format.setSampleSize(SAMPLE_SIZE);
    format.setCodec("audio/pcm");
    format.setByteOrder(QAudioFormat::LittleEndian);
    format.setSampleType(QAudioFormat::SignedInt);

    QAudioDeviceInfo    device = QAudioDeviceInfo::defaultOutputDevice();
    QAudioDeviceInfo info(device);

    if (info.isFormatSupported(format) == false)
        format = info.nearestFormat(format);

    mAudioOutput = new QAudioOutput(device, format);
    mPushTimer = new QTimer();
    connect(mPushTimer, SIGNAL(timeout()), SLOT(pushTimerExpired()));

    mOutput = mAudioOutput->start();
    mPushTimer->start(TIMER_PERIOD);
}

//------------------------------------------------------------
CSoundManager::~CSoundManager()
{
    mPushTimer->stop();
    mAudioOutput->stop();
}

//------------------------------------------------------------
void CSoundManager::pushTimerExpired()
{
    if (mAudioOutput && mAudioOutput->state() != QAudio::StoppedState)
    {
        int len = SAMPLE_BYTES * CHANNELS * SAMPLE_RATE * TIMER_PERIOD / 1000;

        // efface le buffer son
        memset(mBuffer, 0, len);

        if (interface.isGamePaused() == false)
        {
            // trie les sons
            std::vector<CSound*>    sorted;
            for (size_t i = 0; i < mSounds.size(); ++i)
                sorted.push_back(&mSounds[i]);

            for (int i = 0; i < (int)mSounds.size() - 1; ++i)
                for (int j = i + 1; j < (int)mSounds.size(); ++j)
                    if (sorted[j]->getDistance() < sorted[i]->getDistance())
                    {
                        CSound* temp = sorted[i];
                        sorted[i] = sorted[j];
                        sorted[j] = temp;
                    }

            // joue les 4 sons les plus proches (on saute les autres samples)
            int leftVolume, rightVolume;

            for (size_t i = 0; i < sorted.size(); ++i)
            {
                if (i < MIXER_COEF)
                {
                    sorted[i]->getVolumes(&leftVolume, &rightVolume);
                    sorted[i]->getData(mBuffer, len, leftVolume, rightVolume);
                }
                else
                {
                    sorted[i]->skipSamples(len);
                }
            }
        }

        mOutput->write((char *)mBuffer, len);

        // efface les sons terminés
        std::vector<CSound>::iterator   it;
        for (it = mSounds.begin(); it != mSounds.end();)
        {
            if (it->isFinished() == true)
                it = mSounds.erase(it);
            else
                it++;
        }
    }
}

//------------------------------------------------------------
void CSoundManager::play(CVec2 pos, const char* fileName, int loops)
{
    std::vector<CSound>::iterator   it;

    for (it = mSounds.begin(); it !=mSounds.end(); ++it)
        if (it->mPos == pos && it->mFileName == fileName)
            break;

    if (it == mSounds.end())
    {
        CSound  snd(pos, fileName, loops);
        mSounds.push_back(snd);
    }
}

//------------------------------------------------------------
void CSoundManager::stop(CVec2 pos, const char *fileName)
{
    std::vector<CSound>::iterator   it;

    for (it = mSounds.begin(); it !=mSounds.end();)
    {
        if (it->mPos == pos && it->mFileName == fileName)
            it = mSounds.erase(it);
        else
            it++;
    }
}

//------------------------------------------------------------
CSound::CSound(CVec2 pos, const char *fileName, int loops)
{
    mPos = pos;
    mFileName = fileName;
    mLoops = loops;
    mCurrentPos = 0;
    fileCache.getFile(fileName);
}

//------------------------------------------------------------
bool CSound::isFinished()
{
    uint8_t*    mem = (uint8_t*)fileCache.getFile(mFileName).mem;
    uint32_t    size = *(uint32_t*)&mem[WAV_HEADER - 4];

    if (mLoops == 0 || mCurrentPos < mLoops * size)
        return false;

    return true;
}

//------------------------------------------------------------
uint8_t CSound::getSample()
{
    uint8_t*    mem = (uint8_t*)fileCache.getFile(mFileName).mem;
    uint32_t    size = *(uint32_t*)&mem[WAV_HEADER - 4];

    if (isFinished() == false)
        return mem[WAV_HEADER + (mCurrentPos % size)];
    else
        return 128;
}

//------------------------------------------------------------
void CSound::getData(int16_t* dest, int len, int leftVolume, int rightVolume)
{
    for (int i = 0; i < len / (SAMPLE_BYTES * CHANNELS); ++i)
    {
        int sample = (int)getSample() - 128;
        *dest++ += sample * leftVolume / MIXER_COEF;
        *dest++ += sample * rightVolume / MIXER_COEF;
        mCurrentPos++;
    }
}

//------------------------------------------------------------
float CSound::getDistance()
{
    CVec2   pos = player.getLocalFromPos(mPos);
    return pos.length();
}

//------------------------------------------------------------
void CSound::skipSamples(int len)
{
    mCurrentPos += len;
}

//------------------------------------------------------------
void CSound::getVolumes(int* left, int* right)
{
    CVec2   pos = player.getLocalFromPos(mPos);

    if (pos == CVec2(0, 0))
    {
        *left = 256;
        *right = 256;
        return;
    }

    // calcule la balance
    float   angle = atan2(-pos.y, pos.x);

    float l, r;

    if (pos.x < 0)
    {
        l = 1.0f;
        r = fabs(sin(angle));
    }
    else
    {
        l = fabs(sin(angle));
        r = 1.0f;
    }

    // calcule la distance
    float vol = (5.0f - pos.length()) / 5.0f;
    if (vol < 0.0f)
        vol = 0.0f;

    l *= vol;
    r *= vol;

    // retourne les valeurs
    *left = (int)(l * 256.0f);
    *right = (int)(r * 256.0f);
}
